/*
 * This program establishes a DDE conversation with Netscape.  Activeates it,
 * then switches it to the URL specified on the command line or
 * http://www.visix.com/ if none is specified
 */

#define INCL_WIN
#define INCL_DOS
#define INCL_BASE
#include <os2.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>


char *url;
HAB  hab;
HWND netscapeWnd;
int  iRC = 0 ;

void displayErr(char *msg) {
  WinMessageBox(HWND_DESKTOP, NULLHANDLE, msg, "Error", 0, MB_OK);
}

void displayWinErr(char *msg) {
  char theString[1024];

  sprintf(
    theString,
    "%s\nWinGetLastError returned 0x%x",
    msg, WinGetLastError(hab)
  );
  displayErr(theString);
}

void displayMessage(char *msg) {
  WinMessageBox(HWND_DESKTOP, NULLHANDLE, msg, "Message", 0, MB_OK);
}

int winPrintf(const char *formatStr, ...) {
  va_list arg_ptr;

  char buffer[1000];
  int  len;

  va_start(arg_ptr, formatStr);
  len = vsprintf(buffer, formatStr, arg_ptr);
  va_end(arg_ptr);

  displayMessage(buffer);

  return len ;
}

PDDESTRUCT makeDDEObject(
  HWND hwnd, PSZ item, USHORT status, USHORT format,
  PVOID data, USHORT dataLen
) {
  PDDESTRUCT pddes;
  ULONG      itemLen;
  PULONG     pulSharedObj;
  PID        pid;
  TID        tid;

  if (item != NULL) {
    itemLen = strlen(item) + 1;
  } else {
    itemLen = 1;
  }
  
  if (!(DosAllocSharedMem((PPVOID)&pulSharedObj, NULL,
       sizeof(DDESTRUCT) + itemLen + dataLen,
       PAG_COMMIT | PAG_READ | PAG_WRITE |
       OBJ_GIVEABLE)))
  {
    WinQueryWindowProcess(hwnd, &pid, &tid);
    DosGiveSharedMem(&pulSharedObj, pid, PAG_READ | PAG_WRITE);

    pddes = (PDDESTRUCT)pulSharedObj;
    pddes->cbData = (LONG)dataLen;
    pddes->fsStatus = status;
    pddes->usFormat = format;
    pddes->offszItemName = sizeof(DDESTRUCT);
    if ((dataLen) && (data != NULL))
      pddes->offabData = sizeof(DDESTRUCT) + itemLen;
    else
      pddes->offabData = 0;
    
    if (item != NULL)
      strcpy(DDES_PSZITEMNAME(pddes), item);
    else
      strcpy(DDES_PSZITEMNAME(pddes), "");
    
    if (data != NULL)
      memcpy(DDES_PABDATA(pddes), data, dataLen);
    
    return pddes;
  }

  return NULL;
}

/*
 * Send a REQUEST message to the application.  Ignore any return codes
 */
void sendDDEMessage(
  char *application, char *topic, char *param, HWND notify
) {
  BOOL        rc;
  CONVCONTEXT context;
  PDDESTRUCT  pddes;
    
  /* I would mutex lock here, to protect the global hwnd variable */
  
  netscapeWnd = NULLHANDLE;
  memset(&context, 0, sizeof(context));
  context.cb = sizeof(context);
  rc = WinDdeInitiate(notify, application, topic, &context);
  if (rc == FALSE) {
    displayWinErr("WinDdeInitiate returned FALSE");
    return;
  }

  if (netscapeWnd == NULLHANDLE) {
    /* displayErr("WinDdeInitiate didn't find any processes"); */
      
    iRC = 105 ;
    return;
  }

  pddes = makeDDEObject(netscapeWnd, param, 0, DDEFMT_TEXT, NULL, 0);
  rc= WinDdePostMsg(netscapeWnd, notify, WM_DDE_REQUEST, pddes, DDEPM_RETRY);
  if (rc == FALSE)
    displayWinErr("WinDdePostMsg returned FALSE");

  rc = WinDdePostMsg(
    netscapeWnd, notify, WM_DDE_TERMINATE, NULL,
    DDEPM_RETRY
  );
  if (rc == FALSE)
    displayWinErr("WinDdepostMsg(terminate) returned FALSE");
}


/*
 * The main guts of the app
 */
void letsGo(HWND hwnd) {
  char urlString[1000];

  sendDDEMessage("NETSCAPE", "WWW_Activate", "0xFFFFFFFF,0x0", hwnd);

  sprintf(urlString, "%s,,0xFFFFFFFF,0x0,,,", url);
  sendDDEMessage("NETSCAPE", "WWW_OpenURL", urlString, hwnd);

  WinPostMsg(hwnd, WM_QUIT, 0, 0);
}

MRESULT EXPENTRY ClientWindowProc(
  HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2
) {
  char theString[255];
    
  switch(msg) {
    case WM_USER:
      letsGo(hwnd);
    return 0;
    
    case WM_DDE_INITIATEACK: {
      PDDEINIT pddeinit;
      pddeinit = (PDDEINIT)PVOIDFROMMP(mp2);
      printf(
	"Got WM_DDE_INITIATEACK.  hwnd=0x%x, app=%s, topic=%s\n",
	LONGFROMMP(mp1), pddeinit->pszAppName,
	pddeinit->pszTopic
      );
      netscapeWnd = LONGFROMMP(mp1);
    }
    return MRFROMLONG(TRUE);
    
    case WM_DDE_DATA: {
      PDDESTRUCT pddes;
      APIRET     rc;
      pddes = (PDDESTRUCT)PVOIDFROMMP(mp2);
      printf(
	"Got WM_DDE_DATA."
	"hwnd=0x%x, status=0x%x, format=0x%x, item=%s, data=%s\n",
	LONGFROMMP(mp1), pddes->fsStatus, pddes->usFormat,
	(char *)pddes + pddes->offszItemName,
	(char *)pddes + pddes->offabData
      );
      if (pddes->fsStatus & DDE_FACKREQ) {
	/*
	 * We must acknowledge this
	 */
	PDDESTRUCT pddes2 = makeDDEObject(
	  LONGFROMMP(mp1),
	  (char *)pddes + pddes->offszItemName,
	  DDE_FACK | DDE_FRESPONSE,
	  DDEFMT_TEXT, NULL, 0
	) ;
	WinDdePostMsg(
	  LONGFROMMP(mp1), hwnd, WM_DDE_ACK, pddes2,
	  DDEPM_RETRY
	);
      }

      rc = DosFreeMem(pddes);
      if (rc != 0)
	winPrintf(
	  "DosFreeMem(0x%p) returned %d = 0x%x", pddes, rc, rc
	);
      return 0;
    }
    
    default:
    return WinDefWindowProc(hwnd, msg, mp1, mp2);
  }

  return WinDefWindowProc(hwnd, msg, mp1, mp2);
}

int main(int argc, char *argv[]) {
  HMQ hmq;
  HWND hwndTarget;
  QMSG qmsg;
  ULONG flCreate;

  if (argc < 2)
    url = "http://www.visix.com";
  else
    url = argv[1];

  /*
   * Set up and create a window for the DDE response messages to go to
   */
  hab = WinInitialize(0);
  if (hab == NULLHANDLE) {
    printf("WinInitialize failed.\n");
    return 100;
  }

  hmq = WinCreateMsgQueue(hab, 0);
  if (hmq == NULLHANDLE) {
    printf(
      "WinCreatemsgQueue failed.  Err = 0x%x\n",
      WinGetLastError(hab)
    );
    WinTerminate(hab);
    return 101;
  }

  if (!WinRegisterClass(
    hab, "ddensClass", ClientWindowProc, CS_SIZEREDRAW, 0)
  ) {
    displayWinErr("WinRegisterClass(wsaClass) failed.");
    WinDestroyMsgQueue(hmq);
    WinTerminate(hab);
    return 102;
  }

  hwndTarget = WinCreateWindow(
    HWND_OBJECT,  /* parent window */
    "ddensClass", /* window class */
    NULL,         /* window text */
    0,            /* window style */
    0, 0, 0, 0,   /* window position */
    NULLHANDLE,   /* owner window */
    HWND_BOTTOM,  /* Z-order */
    0,            /* window ID */
    NULL,         /* control data */
    NULL        /* presentation params */
  ) ;
  
  if (hwndTarget == NULLHANDLE) {
    displayWinErr("WinCreateWindow failed.");
    WinDestroyMsgQueue(hmq);
    WinTerminate(hab);
    return 103;
  }

  /*
   * Post myself a WM_USER message to start everything.  We don't want to
   * just make calls here, because I want the message loop running.
   */
  if (WinPostMsg(hwndTarget, WM_USER, 0, 0) == FALSE) {
    displayWinErr("Could not WinPostMsg the starting message");
    WinDestroyWindow(hwndTarget);
    WinDestroyMsgQueue(hmq);
    WinTerminate(hab);
    return 104;
  }

  /* The message loop */
  while (WinGetMsg(hab, &qmsg, 0L, 0, 0))
    WinDispatchMsg(hab, &qmsg);

  /* Clean up everything */
  WinDestroyWindow(hwndTarget);
  WinDestroyMsgQueue(hmq);
  WinTerminate(hab);
  
  return iRC ;
}
